#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 16
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
/* Need to handle BETA=0 case by assigning y to zero outside of loop */
void ATL_UGEMV(ATL_CINT M, ATL_CINT N, const TYPE *A, ATL_CINT lda,
               const TYPE *X, TYPE *Y)
/*
 * 24x16 unrolled mvn_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N16=(N/16)*16, M24=(M/24)*24, lda16=lda*16;
   ATL_INT j;

   #ifdef BETA0
      for (j=0; j < M; j++)
         Y[j] = ATL_rzero;
   #endif
   for (j=N16; j; j -= 16, A += lda16, X += 16)
   {
      const double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda,
                   *A5=A4+lda, *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda,
                   *A10=A9+lda, *A11=A10+lda, *A12=A11+lda, *A13=A12+lda,
                   *A14=A13+lda, *A15=A14+lda;
      const register double x0=X[0], x1=X[1], x2=X[2], x3=X[3], x4=X[4],
                            x5=X[5], x6=X[6], x7=X[7], x8=X[8], x9=X[9],
                            x10=X[10], x11=X[11], x12=X[12], x13=X[13],
                            x14=X[14], x15=X[15];
      ATL_INT i;
      prefY(Y+16+16-1);
      for (i=0; i < M24; i += 24)
      {
         const double a0_0=A0[i+0], a1_0=A0[i+1], a2_0=A0[i+2], a3_0=A0[i+3],
                      a4_0=A0[i+4], a5_0=A0[i+5], a6_0=A0[i+6], a7_0=A0[i+7],
                      a8_0=A0[i+8], a9_0=A0[i+9], a10_0=A0[i+10],
                      a11_0=A0[i+11], a12_0=A0[i+12], a13_0=A0[i+13],
                      a14_0=A0[i+14], a15_0=A0[i+15], a16_0=A0[i+16],
                      a17_0=A0[i+17], a18_0=A0[i+18], a19_0=A0[i+19],
                      a20_0=A0[i+20], a21_0=A0[i+21], a22_0=A0[i+22],
                      a23_0=A0[i+23], a0_1=A1[i+0], a1_1=A1[i+1], a2_1=A1[i+2],
                      a3_1=A1[i+3], a4_1=A1[i+4], a5_1=A1[i+5], a6_1=A1[i+6],
                      a7_1=A1[i+7], a8_1=A1[i+8], a9_1=A1[i+9], a10_1=A1[i+10],
                      a11_1=A1[i+11], a12_1=A1[i+12], a13_1=A1[i+13],
                      a14_1=A1[i+14], a15_1=A1[i+15], a16_1=A1[i+16],
                      a17_1=A1[i+17], a18_1=A1[i+18], a19_1=A1[i+19],
                      a20_1=A1[i+20], a21_1=A1[i+21], a22_1=A1[i+22],
                      a23_1=A1[i+23], a0_2=A2[i+0], a1_2=A2[i+1], a2_2=A2[i+2],
                      a3_2=A2[i+3], a4_2=A2[i+4], a5_2=A2[i+5], a6_2=A2[i+6],
                      a7_2=A2[i+7], a8_2=A2[i+8], a9_2=A2[i+9], a10_2=A2[i+10],
                      a11_2=A2[i+11], a12_2=A2[i+12], a13_2=A2[i+13],
                      a14_2=A2[i+14], a15_2=A2[i+15], a16_2=A2[i+16],
                      a17_2=A2[i+17], a18_2=A2[i+18], a19_2=A2[i+19],
                      a20_2=A2[i+20], a21_2=A2[i+21], a22_2=A2[i+22],
                      a23_2=A2[i+23], a0_3=A3[i+0], a1_3=A3[i+1], a2_3=A3[i+2],
                      a3_3=A3[i+3], a4_3=A3[i+4], a5_3=A3[i+5], a6_3=A3[i+6],
                      a7_3=A3[i+7], a8_3=A3[i+8], a9_3=A3[i+9], a10_3=A3[i+10],
                      a11_3=A3[i+11], a12_3=A3[i+12], a13_3=A3[i+13],
                      a14_3=A3[i+14], a15_3=A3[i+15], a16_3=A3[i+16],
                      a17_3=A3[i+17], a18_3=A3[i+18], a19_3=A3[i+19],
                      a20_3=A3[i+20], a21_3=A3[i+21], a22_3=A3[i+22],
                      a23_3=A3[i+23], a0_4=A4[i+0], a1_4=A4[i+1], a2_4=A4[i+2],
                      a3_4=A4[i+3], a4_4=A4[i+4], a5_4=A4[i+5], a6_4=A4[i+6],
                      a7_4=A4[i+7], a8_4=A4[i+8], a9_4=A4[i+9], a10_4=A4[i+10],
                      a11_4=A4[i+11], a12_4=A4[i+12], a13_4=A4[i+13],
                      a14_4=A4[i+14], a15_4=A4[i+15], a16_4=A4[i+16],
                      a17_4=A4[i+17], a18_4=A4[i+18], a19_4=A4[i+19],
                      a20_4=A4[i+20], a21_4=A4[i+21], a22_4=A4[i+22],
                      a23_4=A4[i+23], a0_5=A5[i+0], a1_5=A5[i+1], a2_5=A5[i+2],
                      a3_5=A5[i+3], a4_5=A5[i+4], a5_5=A5[i+5], a6_5=A5[i+6],
                      a7_5=A5[i+7], a8_5=A5[i+8], a9_5=A5[i+9], a10_5=A5[i+10],
                      a11_5=A5[i+11], a12_5=A5[i+12], a13_5=A5[i+13],
                      a14_5=A5[i+14], a15_5=A5[i+15], a16_5=A5[i+16],
                      a17_5=A5[i+17], a18_5=A5[i+18], a19_5=A5[i+19],
                      a20_5=A5[i+20], a21_5=A5[i+21], a22_5=A5[i+22],
                      a23_5=A5[i+23], a0_6=A6[i+0], a1_6=A6[i+1], a2_6=A6[i+2],
                      a3_6=A6[i+3], a4_6=A6[i+4], a5_6=A6[i+5], a6_6=A6[i+6],
                      a7_6=A6[i+7], a8_6=A6[i+8], a9_6=A6[i+9], a10_6=A6[i+10],
                      a11_6=A6[i+11], a12_6=A6[i+12], a13_6=A6[i+13],
                      a14_6=A6[i+14], a15_6=A6[i+15], a16_6=A6[i+16],
                      a17_6=A6[i+17], a18_6=A6[i+18], a19_6=A6[i+19],
                      a20_6=A6[i+20], a21_6=A6[i+21], a22_6=A6[i+22],
                      a23_6=A6[i+23], a0_7=A7[i+0], a1_7=A7[i+1], a2_7=A7[i+2],
                      a3_7=A7[i+3], a4_7=A7[i+4], a5_7=A7[i+5], a6_7=A7[i+6],
                      a7_7=A7[i+7], a8_7=A7[i+8], a9_7=A7[i+9], a10_7=A7[i+10],
                      a11_7=A7[i+11], a12_7=A7[i+12], a13_7=A7[i+13],
                      a14_7=A7[i+14], a15_7=A7[i+15], a16_7=A7[i+16],
                      a17_7=A7[i+17], a18_7=A7[i+18], a19_7=A7[i+19],
                      a20_7=A7[i+20], a21_7=A7[i+21], a22_7=A7[i+22],
                      a23_7=A7[i+23], a0_8=A8[i+0], a1_8=A8[i+1], a2_8=A8[i+2],
                      a3_8=A8[i+3], a4_8=A8[i+4], a5_8=A8[i+5], a6_8=A8[i+6],
                      a7_8=A8[i+7], a8_8=A8[i+8], a9_8=A8[i+9], a10_8=A8[i+10],
                      a11_8=A8[i+11], a12_8=A8[i+12], a13_8=A8[i+13],
                      a14_8=A8[i+14], a15_8=A8[i+15], a16_8=A8[i+16],
                      a17_8=A8[i+17], a18_8=A8[i+18], a19_8=A8[i+19],
                      a20_8=A8[i+20], a21_8=A8[i+21], a22_8=A8[i+22],
                      a23_8=A8[i+23], a0_9=A9[i+0], a1_9=A9[i+1], a2_9=A9[i+2],
                      a3_9=A9[i+3], a4_9=A9[i+4], a5_9=A9[i+5], a6_9=A9[i+6],
                      a7_9=A9[i+7], a8_9=A9[i+8], a9_9=A9[i+9], a10_9=A9[i+10],
                      a11_9=A9[i+11], a12_9=A9[i+12], a13_9=A9[i+13],
                      a14_9=A9[i+14], a15_9=A9[i+15], a16_9=A9[i+16],
                      a17_9=A9[i+17], a18_9=A9[i+18], a19_9=A9[i+19],
                      a20_9=A9[i+20], a21_9=A9[i+21], a22_9=A9[i+22],
                      a23_9=A9[i+23], a0_10=A10[i+0], a1_10=A10[i+1],
                      a2_10=A10[i+2], a3_10=A10[i+3], a4_10=A10[i+4],
                      a5_10=A10[i+5], a6_10=A10[i+6], a7_10=A10[i+7],
                      a8_10=A10[i+8], a9_10=A10[i+9], a10_10=A10[i+10],
                      a11_10=A10[i+11], a12_10=A10[i+12], a13_10=A10[i+13],
                      a14_10=A10[i+14], a15_10=A10[i+15], a16_10=A10[i+16],
                      a17_10=A10[i+17], a18_10=A10[i+18], a19_10=A10[i+19],
                      a20_10=A10[i+20], a21_10=A10[i+21], a22_10=A10[i+22],
                      a23_10=A10[i+23], a0_11=A11[i+0], a1_11=A11[i+1],
                      a2_11=A11[i+2], a3_11=A11[i+3], a4_11=A11[i+4],
                      a5_11=A11[i+5], a6_11=A11[i+6], a7_11=A11[i+7],
                      a8_11=A11[i+8], a9_11=A11[i+9], a10_11=A11[i+10],
                      a11_11=A11[i+11], a12_11=A11[i+12], a13_11=A11[i+13],
                      a14_11=A11[i+14], a15_11=A11[i+15], a16_11=A11[i+16],
                      a17_11=A11[i+17], a18_11=A11[i+18], a19_11=A11[i+19],
                      a20_11=A11[i+20], a21_11=A11[i+21], a22_11=A11[i+22],
                      a23_11=A11[i+23], a0_12=A12[i+0], a1_12=A12[i+1],
                      a2_12=A12[i+2], a3_12=A12[i+3], a4_12=A12[i+4],
                      a5_12=A12[i+5], a6_12=A12[i+6], a7_12=A12[i+7],
                      a8_12=A12[i+8], a9_12=A12[i+9], a10_12=A12[i+10],
                      a11_12=A12[i+11], a12_12=A12[i+12], a13_12=A12[i+13],
                      a14_12=A12[i+14], a15_12=A12[i+15], a16_12=A12[i+16],
                      a17_12=A12[i+17], a18_12=A12[i+18], a19_12=A12[i+19],
                      a20_12=A12[i+20], a21_12=A12[i+21], a22_12=A12[i+22],
                      a23_12=A12[i+23], a0_13=A13[i+0], a1_13=A13[i+1],
                      a2_13=A13[i+2], a3_13=A13[i+3], a4_13=A13[i+4],
                      a5_13=A13[i+5], a6_13=A13[i+6], a7_13=A13[i+7],
                      a8_13=A13[i+8], a9_13=A13[i+9], a10_13=A13[i+10],
                      a11_13=A13[i+11], a12_13=A13[i+12], a13_13=A13[i+13],
                      a14_13=A13[i+14], a15_13=A13[i+15], a16_13=A13[i+16],
                      a17_13=A13[i+17], a18_13=A13[i+18], a19_13=A13[i+19],
                      a20_13=A13[i+20], a21_13=A13[i+21], a22_13=A13[i+22],
                      a23_13=A13[i+23], a0_14=A14[i+0], a1_14=A14[i+1],
                      a2_14=A14[i+2], a3_14=A14[i+3], a4_14=A14[i+4],
                      a5_14=A14[i+5], a6_14=A14[i+6], a7_14=A14[i+7],
                      a8_14=A14[i+8], a9_14=A14[i+9], a10_14=A14[i+10],
                      a11_14=A14[i+11], a12_14=A14[i+12], a13_14=A14[i+13],
                      a14_14=A14[i+14], a15_14=A14[i+15], a16_14=A14[i+16],
                      a17_14=A14[i+17], a18_14=A14[i+18], a19_14=A14[i+19],
                      a20_14=A14[i+20], a21_14=A14[i+21], a22_14=A14[i+22],
                      a23_14=A14[i+23], a0_15=A15[i+0], a1_15=A15[i+1],
                      a2_15=A15[i+2], a3_15=A15[i+3], a4_15=A15[i+4],
                      a5_15=A15[i+5], a6_15=A15[i+6], a7_15=A15[i+7],
                      a8_15=A15[i+8], a9_15=A15[i+9], a10_15=A15[i+10],
                      a11_15=A15[i+11], a12_15=A15[i+12], a13_15=A15[i+13],
                      a14_15=A15[i+14], a15_15=A15[i+15], a16_15=A15[i+16],
                      a17_15=A15[i+17], a18_15=A15[i+18], a19_15=A15[i+19],
                      a20_15=A15[i+20], a21_15=A15[i+21], a22_15=A15[i+22],
                      a23_15=A15[i+23];
         register double y0=Y[i+0], y1=Y[i+1], y2=Y[i+2], y3=Y[i+3], y4=Y[i+4],
                         y5=Y[i+5], y6=Y[i+6], y7=Y[i+7], y8=Y[i+8], y9=Y[i+9],
                         y10=Y[i+10], y11=Y[i+11], y12=Y[i+12], y13=Y[i+13],
                         y14=Y[i+14], y15=Y[i+15], y16=Y[i+16], y17=Y[i+17],
                         y18=Y[i+18], y19=Y[i+19], y20=Y[i+20], y21=Y[i+21],
                         y22=Y[i+22], y23=Y[i+23];

         prefX(X);
         y0 += a0_0 * x0;
         prefA(A0);
         y1 += a1_0 * x0;
         y2 += a2_0 * x0;
         y3 += a3_0 * x0;
         y4 += a4_0 * x0;
         y5 += a5_0 * x0;
         y6 += a6_0 * x0;
         y7 += a7_0 * x0;
         y8 += a8_0 * x0;
         prefA(A0);
         y9 += a9_0 * x0;
         y10 += a10_0 * x0;
         y11 += a11_0 * x0;
         y12 += a12_0 * x0;
         y13 += a13_0 * x0;
         y14 += a14_0 * x0;
         y15 += a15_0 * x0;
         y16 += a16_0 * x0;
         prefA(A0);
         y17 += a17_0 * x0;
         y18 += a18_0 * x0;
         y19 += a19_0 * x0;
         y20 += a20_0 * x0;
         y21 += a21_0 * x0;
         y22 += a22_0 * x0;
         y23 += a23_0 * x0;
         y0 += a0_1 * x1;
         prefA(A1);
         y1 += a1_1 * x1;
         y2 += a2_1 * x1;
         y3 += a3_1 * x1;
         y4 += a4_1 * x1;
         y5 += a5_1 * x1;
         y6 += a6_1 * x1;
         y7 += a7_1 * x1;
         y8 += a8_1 * x1;
         prefA(A1);
         y9 += a9_1 * x1;
         y10 += a10_1 * x1;
         y11 += a11_1 * x1;
         y12 += a12_1 * x1;
         y13 += a13_1 * x1;
         y14 += a14_1 * x1;
         y15 += a15_1 * x1;
         y16 += a16_1 * x1;
         prefA(A1);
         y17 += a17_1 * x1;
         y18 += a18_1 * x1;
         y19 += a19_1 * x1;
         y20 += a20_1 * x1;
         y21 += a21_1 * x1;
         y22 += a22_1 * x1;
         y23 += a23_1 * x1;
         y0 += a0_2 * x2;
         prefA(A2);
         y1 += a1_2 * x2;
         y2 += a2_2 * x2;
         y3 += a3_2 * x2;
         y4 += a4_2 * x2;
         y5 += a5_2 * x2;
         y6 += a6_2 * x2;
         y7 += a7_2 * x2;
         y8 += a8_2 * x2;
         prefA(A2);
         y9 += a9_2 * x2;
         y10 += a10_2 * x2;
         y11 += a11_2 * x2;
         y12 += a12_2 * x2;
         y13 += a13_2 * x2;
         y14 += a14_2 * x2;
         y15 += a15_2 * x2;
         y16 += a16_2 * x2;
         prefA(A2);
         y17 += a17_2 * x2;
         y18 += a18_2 * x2;
         y19 += a19_2 * x2;
         y20 += a20_2 * x2;
         y21 += a21_2 * x2;
         y22 += a22_2 * x2;
         y23 += a23_2 * x2;
         y0 += a0_3 * x3;
         prefA(A3);
         y1 += a1_3 * x3;
         y2 += a2_3 * x3;
         y3 += a3_3 * x3;
         y4 += a4_3 * x3;
         y5 += a5_3 * x3;
         y6 += a6_3 * x3;
         y7 += a7_3 * x3;
         y8 += a8_3 * x3;
         prefA(A3);
         y9 += a9_3 * x3;
         y10 += a10_3 * x3;
         y11 += a11_3 * x3;
         y12 += a12_3 * x3;
         y13 += a13_3 * x3;
         y14 += a14_3 * x3;
         y15 += a15_3 * x3;
         y16 += a16_3 * x3;
         prefA(A3);
         y17 += a17_3 * x3;
         y18 += a18_3 * x3;
         y19 += a19_3 * x3;
         y20 += a20_3 * x3;
         y21 += a21_3 * x3;
         y22 += a22_3 * x3;
         y23 += a23_3 * x3;
         y0 += a0_4 * x4;
         prefA(A4);
         y1 += a1_4 * x4;
         y2 += a2_4 * x4;
         y3 += a3_4 * x4;
         y4 += a4_4 * x4;
         y5 += a5_4 * x4;
         y6 += a6_4 * x4;
         y7 += a7_4 * x4;
         y8 += a8_4 * x4;
         prefA(A4);
         y9 += a9_4 * x4;
         y10 += a10_4 * x4;
         y11 += a11_4 * x4;
         y12 += a12_4 * x4;
         y13 += a13_4 * x4;
         y14 += a14_4 * x4;
         y15 += a15_4 * x4;
         y16 += a16_4 * x4;
         prefA(A4);
         y17 += a17_4 * x4;
         y18 += a18_4 * x4;
         y19 += a19_4 * x4;
         y20 += a20_4 * x4;
         y21 += a21_4 * x4;
         y22 += a22_4 * x4;
         y23 += a23_4 * x4;
         y0 += a0_5 * x5;
         prefA(A5);
         y1 += a1_5 * x5;
         y2 += a2_5 * x5;
         y3 += a3_5 * x5;
         y4 += a4_5 * x5;
         y5 += a5_5 * x5;
         y6 += a6_5 * x5;
         y7 += a7_5 * x5;
         y8 += a8_5 * x5;
         prefA(A5);
         y9 += a9_5 * x5;
         y10 += a10_5 * x5;
         y11 += a11_5 * x5;
         y12 += a12_5 * x5;
         y13 += a13_5 * x5;
         y14 += a14_5 * x5;
         y15 += a15_5 * x5;
         y16 += a16_5 * x5;
         prefA(A5);
         y17 += a17_5 * x5;
         y18 += a18_5 * x5;
         y19 += a19_5 * x5;
         y20 += a20_5 * x5;
         y21 += a21_5 * x5;
         y22 += a22_5 * x5;
         y23 += a23_5 * x5;
         y0 += a0_6 * x6;
         prefA(A6);
         y1 += a1_6 * x6;
         y2 += a2_6 * x6;
         y3 += a3_6 * x6;
         y4 += a4_6 * x6;
         y5 += a5_6 * x6;
         y6 += a6_6 * x6;
         y7 += a7_6 * x6;
         y8 += a8_6 * x6;
         prefA(A6);
         y9 += a9_6 * x6;
         y10 += a10_6 * x6;
         y11 += a11_6 * x6;
         y12 += a12_6 * x6;
         y13 += a13_6 * x6;
         y14 += a14_6 * x6;
         y15 += a15_6 * x6;
         y16 += a16_6 * x6;
         prefA(A6);
         y17 += a17_6 * x6;
         y18 += a18_6 * x6;
         y19 += a19_6 * x6;
         y20 += a20_6 * x6;
         y21 += a21_6 * x6;
         y22 += a22_6 * x6;
         y23 += a23_6 * x6;
         y0 += a0_7 * x7;
         prefA(A7);
         y1 += a1_7 * x7;
         y2 += a2_7 * x7;
         y3 += a3_7 * x7;
         y4 += a4_7 * x7;
         y5 += a5_7 * x7;
         y6 += a6_7 * x7;
         y7 += a7_7 * x7;
         y8 += a8_7 * x7;
         prefA(A7);
         y9 += a9_7 * x7;
         y10 += a10_7 * x7;
         y11 += a11_7 * x7;
         y12 += a12_7 * x7;
         y13 += a13_7 * x7;
         y14 += a14_7 * x7;
         y15 += a15_7 * x7;
         y16 += a16_7 * x7;
         prefA(A7);
         y17 += a17_7 * x7;
         y18 += a18_7 * x7;
         y19 += a19_7 * x7;
         y20 += a20_7 * x7;
         y21 += a21_7 * x7;
         y22 += a22_7 * x7;
         y23 += a23_7 * x7;
         y0 += a0_8 * x8;
         prefA(A8);
         y1 += a1_8 * x8;
         y2 += a2_8 * x8;
         y3 += a3_8 * x8;
         y4 += a4_8 * x8;
         y5 += a5_8 * x8;
         y6 += a6_8 * x8;
         y7 += a7_8 * x8;
         y8 += a8_8 * x8;
         prefA(A8);
         y9 += a9_8 * x8;
         y10 += a10_8 * x8;
         y11 += a11_8 * x8;
         y12 += a12_8 * x8;
         y13 += a13_8 * x8;
         y14 += a14_8 * x8;
         y15 += a15_8 * x8;
         y16 += a16_8 * x8;
         prefA(A8);
         y17 += a17_8 * x8;
         y18 += a18_8 * x8;
         y19 += a19_8 * x8;
         y20 += a20_8 * x8;
         y21 += a21_8 * x8;
         y22 += a22_8 * x8;
         y23 += a23_8 * x8;
         y0 += a0_9 * x9;
         prefA(A9);
         y1 += a1_9 * x9;
         y2 += a2_9 * x9;
         y3 += a3_9 * x9;
         y4 += a4_9 * x9;
         y5 += a5_9 * x9;
         y6 += a6_9 * x9;
         y7 += a7_9 * x9;
         y8 += a8_9 * x9;
         prefA(A9);
         y9 += a9_9 * x9;
         y10 += a10_9 * x9;
         y11 += a11_9 * x9;
         y12 += a12_9 * x9;
         y13 += a13_9 * x9;
         y14 += a14_9 * x9;
         y15 += a15_9 * x9;
         y16 += a16_9 * x9;
         prefA(A9);
         y17 += a17_9 * x9;
         y18 += a18_9 * x9;
         y19 += a19_9 * x9;
         y20 += a20_9 * x9;
         y21 += a21_9 * x9;
         y22 += a22_9 * x9;
         y23 += a23_9 * x9;
         y0 += a0_10 * x10;
         prefA(A10);
         y1 += a1_10 * x10;
         y2 += a2_10 * x10;
         y3 += a3_10 * x10;
         y4 += a4_10 * x10;
         y5 += a5_10 * x10;
         y6 += a6_10 * x10;
         y7 += a7_10 * x10;
         y8 += a8_10 * x10;
         prefA(A10);
         y9 += a9_10 * x10;
         y10 += a10_10 * x10;
         y11 += a11_10 * x10;
         y12 += a12_10 * x10;
         y13 += a13_10 * x10;
         y14 += a14_10 * x10;
         y15 += a15_10 * x10;
         y16 += a16_10 * x10;
         prefA(A10);
         y17 += a17_10 * x10;
         y18 += a18_10 * x10;
         y19 += a19_10 * x10;
         y20 += a20_10 * x10;
         y21 += a21_10 * x10;
         y22 += a22_10 * x10;
         y23 += a23_10 * x10;
         y0 += a0_11 * x11;
         prefA(A11);
         y1 += a1_11 * x11;
         y2 += a2_11 * x11;
         y3 += a3_11 * x11;
         y4 += a4_11 * x11;
         y5 += a5_11 * x11;
         y6 += a6_11 * x11;
         y7 += a7_11 * x11;
         y8 += a8_11 * x11;
         prefA(A11);
         y9 += a9_11 * x11;
         y10 += a10_11 * x11;
         y11 += a11_11 * x11;
         y12 += a12_11 * x11;
         y13 += a13_11 * x11;
         y14 += a14_11 * x11;
         y15 += a15_11 * x11;
         y16 += a16_11 * x11;
         prefA(A11);
         y17 += a17_11 * x11;
         y18 += a18_11 * x11;
         y19 += a19_11 * x11;
         y20 += a20_11 * x11;
         y21 += a21_11 * x11;
         y22 += a22_11 * x11;
         y23 += a23_11 * x11;
         y0 += a0_12 * x12;
         prefA(A12);
         y1 += a1_12 * x12;
         y2 += a2_12 * x12;
         y3 += a3_12 * x12;
         y4 += a4_12 * x12;
         y5 += a5_12 * x12;
         y6 += a6_12 * x12;
         y7 += a7_12 * x12;
         y8 += a8_12 * x12;
         prefA(A12);
         y9 += a9_12 * x12;
         y10 += a10_12 * x12;
         y11 += a11_12 * x12;
         y12 += a12_12 * x12;
         y13 += a13_12 * x12;
         y14 += a14_12 * x12;
         y15 += a15_12 * x12;
         y16 += a16_12 * x12;
         prefA(A12);
         y17 += a17_12 * x12;
         y18 += a18_12 * x12;
         y19 += a19_12 * x12;
         y20 += a20_12 * x12;
         y21 += a21_12 * x12;
         y22 += a22_12 * x12;
         y23 += a23_12 * x12;
         y0 += a0_13 * x13;
         prefA(A13);
         y1 += a1_13 * x13;
         y2 += a2_13 * x13;
         y3 += a3_13 * x13;
         y4 += a4_13 * x13;
         y5 += a5_13 * x13;
         y6 += a6_13 * x13;
         y7 += a7_13 * x13;
         y8 += a8_13 * x13;
         prefA(A13);
         y9 += a9_13 * x13;
         y10 += a10_13 * x13;
         y11 += a11_13 * x13;
         y12 += a12_13 * x13;
         y13 += a13_13 * x13;
         y14 += a14_13 * x13;
         y15 += a15_13 * x13;
         y16 += a16_13 * x13;
         prefA(A13);
         y17 += a17_13 * x13;
         y18 += a18_13 * x13;
         y19 += a19_13 * x13;
         y20 += a20_13 * x13;
         y21 += a21_13 * x13;
         y22 += a22_13 * x13;
         y23 += a23_13 * x13;
         y0 += a0_14 * x14;
         prefA(A14);
         y1 += a1_14 * x14;
         y2 += a2_14 * x14;
         y3 += a3_14 * x14;
         y4 += a4_14 * x14;
         y5 += a5_14 * x14;
         y6 += a6_14 * x14;
         y7 += a7_14 * x14;
         y8 += a8_14 * x14;
         prefA(A14);
         y9 += a9_14 * x14;
         y10 += a10_14 * x14;
         y11 += a11_14 * x14;
         y12 += a12_14 * x14;
         y13 += a13_14 * x14;
         y14 += a14_14 * x14;
         y15 += a15_14 * x14;
         y16 += a16_14 * x14;
         prefA(A14);
         y17 += a17_14 * x14;
         y18 += a18_14 * x14;
         y19 += a19_14 * x14;
         y20 += a20_14 * x14;
         y21 += a21_14 * x14;
         y22 += a22_14 * x14;
         y23 += a23_14 * x14;
         y0 += a0_15 * x15;
         prefA(A15);
         y1 += a1_15 * x15;
         y2 += a2_15 * x15;
         y3 += a3_15 * x15;
         y4 += a4_15 * x15;
         y5 += a5_15 * x15;
         y6 += a6_15 * x15;
         y7 += a7_15 * x15;
         y8 += a8_15 * x15;
         prefA(A15);
         y9 += a9_15 * x15;
         y10 += a10_15 * x15;
         y11 += a11_15 * x15;
         y12 += a12_15 * x15;
         y13 += a13_15 * x15;
         y14 += a14_15 * x15;
         y15 += a15_15 * x15;
         y16 += a16_15 * x15;
         prefA(A15);
         y17 += a17_15 * x15;
         y18 += a18_15 * x15;
         y19 += a19_15 * x15;
         y20 += a20_15 * x15;
         y21 += a21_15 * x15;
         y22 += a22_15 * x15;
         y23 += a23_15 * x15;
         Y[i+0] = y0;
         Y[i+1] = y1;
         Y[i+2] = y2;
         Y[i+3] = y3;
         Y[i+4] = y4;
         Y[i+5] = y5;
         Y[i+6] = y6;
         Y[i+7] = y7;
         Y[i+8] = y8;
         Y[i+9] = y9;
         Y[i+10] = y10;
         Y[i+11] = y11;
         Y[i+12] = y12;
         Y[i+13] = y13;
         Y[i+14] = y14;
         Y[i+15] = y15;
         Y[i+16] = y16;
         Y[i+17] = y17;
         Y[i+18] = y18;
         Y[i+19] = y19;
         Y[i+20] = y20;
         Y[i+21] = y21;
         Y[i+22] = y22;
         Y[i+23] = y23;
      }
      for (i=M24; i < M; i++)
      {
         const double a0_0=A0[i], a0_1=A1[i], a0_2=A2[i], a0_3=A3[i],
                      a0_4=A4[i], a0_5=A5[i], a0_6=A6[i], a0_7=A7[i],
                      a0_8=A8[i], a0_9=A9[i], a0_10=A10[i], a0_11=A11[i],
                      a0_12=A12[i], a0_13=A13[i], a0_14=A14[i], a0_15=A15[i];
         register double y0 = Y[i];

         y0 += a0_0 * x0;
         y0 += a0_1 * x1;
         y0 += a0_2 * x2;
         y0 += a0_3 * x3;
         y0 += a0_4 * x4;
         y0 += a0_5 * x5;
         y0 += a0_6 * x6;
         y0 += a0_7 * x7;
         y0 += a0_8 * x8;
         y0 += a0_9 * x9;
         y0 += a0_10 * x10;
         y0 += a0_11 * x11;
         y0 += a0_12 * x12;
         y0 += a0_13 * x13;
         y0 += a0_14 * x14;
         y0 += a0_15 * x15;
         Y[i] = y0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N16; j; j--, A += lda, X++)
   {
      const double *A0=A;
      const register double x0=X[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M24; i += 24)
      {
         const double a0_0=A0[i+0], a1_0=A0[i+1], a2_0=A0[i+2], a3_0=A0[i+3],
                      a4_0=A0[i+4], a5_0=A0[i+5], a6_0=A0[i+6], a7_0=A0[i+7],
                      a8_0=A0[i+8], a9_0=A0[i+9], a10_0=A0[i+10],
                      a11_0=A0[i+11], a12_0=A0[i+12], a13_0=A0[i+13],
                      a14_0=A0[i+14], a15_0=A0[i+15], a16_0=A0[i+16],
                      a17_0=A0[i+17], a18_0=A0[i+18], a19_0=A0[i+19],
                      a20_0=A0[i+20], a21_0=A0[i+21], a22_0=A0[i+22],
                      a23_0=A0[i+23];
         register double y0=Y[i+0], y1=Y[i+1], y2=Y[i+2], y3=Y[i+3], y4=Y[i+4],
                         y5=Y[i+5], y6=Y[i+6], y7=Y[i+7], y8=Y[i+8], y9=Y[i+9],
                         y10=Y[i+10], y11=Y[i+11], y12=Y[i+12], y13=Y[i+13],
                         y14=Y[i+14], y15=Y[i+15], y16=Y[i+16], y17=Y[i+17],
                         y18=Y[i+18], y19=Y[i+19], y20=Y[i+20], y21=Y[i+21],
                         y22=Y[i+22], y23=Y[i+23];

         prefX(X);
         y0 += a0_0 * x0;
         prefA(A0);
         y1 += a1_0 * x0;
         y2 += a2_0 * x0;
         y3 += a3_0 * x0;
         y4 += a4_0 * x0;
         y5 += a5_0 * x0;
         y6 += a6_0 * x0;
         y7 += a7_0 * x0;
         y8 += a8_0 * x0;
         prefA(A0);
         y9 += a9_0 * x0;
         y10 += a10_0 * x0;
         y11 += a11_0 * x0;
         y12 += a12_0 * x0;
         y13 += a13_0 * x0;
         y14 += a14_0 * x0;
         y15 += a15_0 * x0;
         y16 += a16_0 * x0;
         prefA(A0);
         y17 += a17_0 * x0;
         y18 += a18_0 * x0;
         y19 += a19_0 * x0;
         y20 += a20_0 * x0;
         y21 += a21_0 * x0;
         y22 += a22_0 * x0;
         y23 += a23_0 * x0;
         Y[i+0] = y0;
         Y[i+1] = y1;
         Y[i+2] = y2;
         Y[i+3] = y3;
         Y[i+4] = y4;
         Y[i+5] = y5;
         Y[i+6] = y6;
         Y[i+7] = y7;
         Y[i+8] = y8;
         Y[i+9] = y9;
         Y[i+10] = y10;
         Y[i+11] = y11;
         Y[i+12] = y12;
         Y[i+13] = y13;
         Y[i+14] = y14;
         Y[i+15] = y15;
         Y[i+16] = y16;
         Y[i+17] = y17;
         Y[i+18] = y18;
         Y[i+19] = y19;
         Y[i+20] = y20;
         Y[i+21] = y21;
         Y[i+22] = y22;
         Y[i+23] = y23;
      }
      for (i=M24; i < M; i++)
      {
         const double a0_0=A0[i];
         register double y0 = Y[i];

         y0 += a0_0 * x0;
         Y[i] = y0;
      }
   }
}
