#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 2
#endif
#ifndef PFXDIST
   #define PFXDIST 256
#endif
#ifndef PFADIST
   #define PFADIST 256
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M0, ATL_CINT N, const float *X, const float *Y, 
               float *A, ATL_CINT lda0)
/*
 * 16x2 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N2=(N/2)*2, M=M0+M0, M16=((M0/16)*16)<<1,
            lda=lda0+lda0, lda2=lda*2;
   ATL_INT j;

   for (j=N2; j; j -= 2, A += lda2, Y += 4)
   {
      float *A0=A, *A1=A0+lda;
      const register float ry0=Y[0], iy0=Y[1], ry1=Y[2], iy1=Y[3];
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         const register float rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                              rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                              rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                              rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                              ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                              rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                              ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                              rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                              ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                              rx15=X[i+30], ix15=X[i+31];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A1[i+0] += ry1*rx0 - iy1*ix0;
         prefA(A1);
         A1[i+0+1] += ry1*ix0 + iy1*rx0;
         A0[i+2] += ry0*rx1 - iy0*ix1;
         A0[i+2+1] += ry0*ix1 + iy0*rx1;
         A1[i+2] += ry1*rx1 - iy1*ix1;
         A1[i+2+1] += ry1*ix1 + iy1*rx1;
         A0[i+4] += ry0*rx2 - iy0*ix2;
         A0[i+4+1] += ry0*ix2 + iy0*rx2;
         A1[i+4] += ry1*rx2 - iy1*ix2;
         A1[i+4+1] += ry1*ix2 + iy1*rx2;
         A0[i+6] += ry0*rx3 - iy0*ix3;
         A0[i+6+1] += ry0*ix3 + iy0*rx3;
         A1[i+6] += ry1*rx3 - iy1*ix3;
         A1[i+6+1] += ry1*ix3 + iy1*rx3;
         A0[i+8] += ry0*rx4 - iy0*ix4;
         A0[i+8+1] += ry0*ix4 + iy0*rx4;
         A1[i+8] += ry1*rx4 - iy1*ix4;
         A1[i+8+1] += ry1*ix4 + iy1*rx4;
         A0[i+10] += ry0*rx5 - iy0*ix5;
         A0[i+10+1] += ry0*ix5 + iy0*rx5;
         A1[i+10] += ry1*rx5 - iy1*ix5;
         A1[i+10+1] += ry1*ix5 + iy1*rx5;
         A0[i+12] += ry0*rx6 - iy0*ix6;
         A0[i+12+1] += ry0*ix6 + iy0*rx6;
         A1[i+12] += ry1*rx6 - iy1*ix6;
         A1[i+12+1] += ry1*ix6 + iy1*rx6;
         A0[i+14] += ry0*rx7 - iy0*ix7;
         A0[i+14+1] += ry0*ix7 + iy0*rx7;
         A1[i+14] += ry1*rx7 - iy1*ix7;
         A1[i+14+1] += ry1*ix7 + iy1*rx7;
         A0[i+16] += ry0*rx8 - iy0*ix8;
         prefA(A0);
         A0[i+16+1] += ry0*ix8 + iy0*rx8;
         A1[i+16] += ry1*rx8 - iy1*ix8;
         prefA(A1);
         A1[i+16+1] += ry1*ix8 + iy1*rx8;
         A0[i+18] += ry0*rx9 - iy0*ix9;
         A0[i+18+1] += ry0*ix9 + iy0*rx9;
         A1[i+18] += ry1*rx9 - iy1*ix9;
         A1[i+18+1] += ry1*ix9 + iy1*rx9;
         A0[i+20] += ry0*rx10 - iy0*ix10;
         A0[i+20+1] += ry0*ix10 + iy0*rx10;
         A1[i+20] += ry1*rx10 - iy1*ix10;
         A1[i+20+1] += ry1*ix10 + iy1*rx10;
         A0[i+22] += ry0*rx11 - iy0*ix11;
         A0[i+22+1] += ry0*ix11 + iy0*rx11;
         A1[i+22] += ry1*rx11 - iy1*ix11;
         A1[i+22+1] += ry1*ix11 + iy1*rx11;
         A0[i+24] += ry0*rx12 - iy0*ix12;
         A0[i+24+1] += ry0*ix12 + iy0*rx12;
         A1[i+24] += ry1*rx12 - iy1*ix12;
         A1[i+24+1] += ry1*ix12 + iy1*rx12;
         A0[i+26] += ry0*rx13 - iy0*ix13;
         A0[i+26+1] += ry0*ix13 + iy0*rx13;
         A1[i+26] += ry1*rx13 - iy1*ix13;
         A1[i+26+1] += ry1*ix13 + iy1*rx13;
         A0[i+28] += ry0*rx14 - iy0*ix14;
         A0[i+28+1] += ry0*ix14 + iy0*rx14;
         A1[i+28] += ry1*rx14 - iy1*ix14;
         A1[i+28+1] += ry1*ix14 + iy1*rx14;
         A0[i+30] += ry0*rx15 - iy0*ix15;
         A0[i+30+1] += ry0*ix15 + iy0*rx15;
         A1[i+30] += ry1*rx15 - iy1*ix15;
         A1[i+30+1] += ry1*ix15 + iy1*rx15;
      }
      for (i=M16; i < M; i += 2)
      {
         const register float rx0 = X[i], ix0 = X[i+1];
         A0[i] += ry0 * rx0 - iy0 * ix0;
         A0[i+1] += ry0 * ix0 + iy0 * rx0;
         A1[i] += ry1 * rx0 - iy1 * ix0;
         A1[i+1] += ry1 * ix0 + iy1 * rx0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N2; j; j--, A += lda, Y += 2)
   {
      float *A0=A;
      const register float ry0=Y[0], iy0=Y[1];
      ATL_INT i;
      prefY(Y+4+4-1);
      for (i=0; i < M16; i += 32)
      {
         const register float rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                              rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                              rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                              rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                              ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                              rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                              ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                              rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                              ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                              rx15=X[i+30], ix15=X[i+31];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A0[i+2] += ry0*rx1 - iy0*ix1;
         A0[i+2+1] += ry0*ix1 + iy0*rx1;
         A0[i+4] += ry0*rx2 - iy0*ix2;
         A0[i+4+1] += ry0*ix2 + iy0*rx2;
         A0[i+6] += ry0*rx3 - iy0*ix3;
         A0[i+6+1] += ry0*ix3 + iy0*rx3;
         A0[i+8] += ry0*rx4 - iy0*ix4;
         A0[i+8+1] += ry0*ix4 + iy0*rx4;
         A0[i+10] += ry0*rx5 - iy0*ix5;
         A0[i+10+1] += ry0*ix5 + iy0*rx5;
         A0[i+12] += ry0*rx6 - iy0*ix6;
         A0[i+12+1] += ry0*ix6 + iy0*rx6;
         A0[i+14] += ry0*rx7 - iy0*ix7;
         A0[i+14+1] += ry0*ix7 + iy0*rx7;
         A0[i+16] += ry0*rx8 - iy0*ix8;
         prefA(A0);
         A0[i+16+1] += ry0*ix8 + iy0*rx8;
         A0[i+18] += ry0*rx9 - iy0*ix9;
         A0[i+18+1] += ry0*ix9 + iy0*rx9;
         A0[i+20] += ry0*rx10 - iy0*ix10;
         A0[i+20+1] += ry0*ix10 + iy0*rx10;
         A0[i+22] += ry0*rx11 - iy0*ix11;
         A0[i+22+1] += ry0*ix11 + iy0*rx11;
         A0[i+24] += ry0*rx12 - iy0*ix12;
         A0[i+24+1] += ry0*ix12 + iy0*rx12;
         A0[i+26] += ry0*rx13 - iy0*ix13;
         A0[i+26+1] += ry0*ix13 + iy0*rx13;
         A0[i+28] += ry0*rx14 - iy0*ix14;
         A0[i+28+1] += ry0*ix14 + iy0*rx14;
         A0[i+30] += ry0*rx15 - iy0*ix15;
         A0[i+30+1] += ry0*ix15 + iy0*rx15;
      }
      for (i=M16; i < M; i += 2)
      {
         const register float rx0 = X[i], ix0 = X[i+1];
         A0[i] += ry0 * rx0 - iy0 * ix0;
         A0[i+1] += ry0 * ix0 + iy0 * rx0;
      }
   }
}
